<?php

namespace Drupal\content_first\Controller;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Component\Utility\DeprecationHelper;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Returns Node routes.
 */
class NodeController extends EntityController {

  /**
   * Builds the response.
   */
  public function build($node) {

    $build = [];
    if (!$node instanceof EntityInterface) {
      throw new NotFoundHttpException();
    }

    // Render entity HTML output:
    $view_mode = 'full';
    $view_builder = $this->entityTypeManager->getViewBuilder($node->getEntityTypeId());
    $pre_render = $view_builder->view($node, $view_mode);
    $render_output = DeprecationHelper::backwardsCompatibleCall(
      currentVersion: \Drupal::VERSION,
      deprecatedVersion: '10.3',
      currentCallable: fn() => $this->renderer->renderInIsolation($pre_render),
      deprecatedCallable: fn() => $this->renderer->renderPlain($pre_render),
    );

    // Generate the content:
    // Clean up:
    $allowed_tags = '<h1><h2><h3><h4><h5><h6>';
    $allowed_tags .= '<a><p><br>';
    $allowed_tags .= '<strong><em><b><i><u>';
    $allowed_tags .= '<ul><li><ol>';
    $allowed_tags .= '<table><tr><td><th><tbody><thead><tfoot>';
    $allowed_tags .= '<img>';
    $clean = strip_tags($render_output, $allowed_tags);
    $clean = $this->transformTags($clean, ['img', 'iframe', 'video'], ['alt', 'title']);

    if (!empty($clean)) {
      $clean = (string) $this->addMarkdownToHeadings($clean);
    }

    // Generate headers menu items:
    $headers_items = [];
    // Query all h1 to h6 tags.
    $dom = new \DOMDocument();
    // Suppress warnings due to malformed HTML.
    $html = mb_encode_numericentity($render_output, [0x80, 0x10FFFF, 0, 0x1FFFFF], 'UTF-8');
    @$dom->loadHTML($html);
    $xpath = new \DOMXPath($dom);
    $headers = $xpath->query('//h1 | //h2 | //h3 | //h4 | //h5 | //h6');
    if ($headers) {
      $headers_items = $this->buildHeadersMenuItemsFromDom($headers);
    }

    // Main container:
    $build['container'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => ['content-first'],
      ],
    ];

    // Content container:
    $build['container']['content'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['content'],
      ],
    ];

    $build['container']['content']['button'] = [
      '#type' => 'button',
      '#value' => '🗐 ' . $this->t('Copy'),
      '#attributes' => [
        'id' => ['copy-button'],
      ],
    ];

    $build['container']['content']['html'] = [
      '#type' => 'container',
      '#markup' => $clean,
      '#attributes' => [
        'id' => ['content-first-content'],
      ],
    ];

    // Headers container:
    $build['container']['headers'] = [
      '#theme' => 'item_list',
      '#items' => $headers_items,
      '#list_type' => 'ul',
      '#attributes' => [
        'class' => ['headers-menu'],
      ],
    ];

    $build['#attached'] = [
      'library' => [
        'content_first/content_first',
      ],
    ];

    return $build;
  }

  /**
   * Build a multi-level menu items from a DOMNodeList of heading elements.
   *
   * Each DOMNode is expected to be an element like <h2> or <h3>.
   * We interpret the number in the tagName ("H2" => 2, "H3" => 3, etc.)
   * as the nesting level.
   *
   * The returned structure is an array of items, where each item has:
   *   [
   *     '#markup'  => (string) The text inside the heading,
   *     '#level'   => (int)    The heading level,
   *     'children' => array    Nested sub-items,
   *   ]
   *
   * @param \DOMNodeList $nodeList
   *   A DOMNodeList of heading elements (e.g. <h1>, <h2>, <h3>, etc.).
   *
   * @return array
   *   Nested array of items with '#markup', '#level', 'children'.
   */
  public function buildHeadersMenuItemsFromDom(\DOMNodeList $nodeList) {
    // Top-level items in its 'children'.
    $virtualRoot = [
      '#markup'  => '',
      '#level'   => 0,
      'children' => [],
    ];

    // Stack to track current nesting context.
    // Each stack element: [ 'level' => int, 'item' => &theItem ]
    // Start with the virtual root.
    $stack = [
      [
        'level' => 0,
        'item'  => &$virtualRoot,
      ],
    ];

    // Iterate each heading node in the DOMNodeList.
    foreach ($nodeList as $node) {
      if (!$node instanceof \DOMElement) {
        continue;
      }

      // Skip if it doesn't start with 'H'.
      $tagName = strtoupper($node->tagName);
      if (strpos($tagName, 'H') !== 0) {
        continue;
      }

      // Skip or handle non-numeric or invalid headings.
      $lvl = (int) substr($tagName, 1);
      if ($lvl <= 0) {
        continue;
      }

      // The heading text is the node's textContent:
      $text = "<span class=\"html-tag\">$node->tagName</span> $node->textContent";

      // Climb up the stack while the top of stack has a level >= $lvl.
      while (count($stack) > 0 && $stack[count($stack) - 1]['level'] >= $lvl) {
        array_pop($stack);
      }

      // Now the top of stack is a heading whose level < $lvl (or virtual root).
      if (!empty($stack)) {
        $parentItem = &$stack[count($stack) - 1]['item'];

        // Create a new item array.
        $newItem = [
          '#markup'  => $text,
          '#level'   => $lvl,
          'children' => [],
        ];

        // Append it to the parent's 'children'.
        $parentItem['children'][] = $newItem;

        // Push this new item onto the stack so deeper headings become its
        // children.
        $newIndex = array_key_last($parentItem['children']);
        $stack[] = [
          'level' => $lvl,
          'item'  => &$parentItem['children'][$newIndex],
        ];
      }
    }

    // Return everything under the virtual root (the real items).
    return $virtualRoot['children'];
  }

  /**
   * Transform specified HTML tags into bracketed tokens.
   *
   * @param string $content
   *   The HTML (or text) to process.
   * @param string[] $tags
   *   Array of tag names to match, e.g. ['img','iframe','video'].
   * @param string[] $attributes
   *   Array of attributes to capture, e.g. ['src','alt','title'].
   *
   * @return string
   *   The transformed content
   */
  public function transformTags(string $content, array $tags, array $attributes): string {
    if (empty($content)) {
      return "";
    }

    // Build a regex like /(<(img|iframe|video)[^>]*>)/i.
    $tagPattern = implode('|', array_map('preg_quote', $tags));
    $pattern = '/<(' . $tagPattern . ')[^>]*>/i';

    $content = preg_replace_callback($pattern, function ($matches) use ($attributes) {
      // e.g. <img src="..." alt="..." />.
      $fullTag = $matches[0];
      // e.g. 'img'.
      $tagName = strtolower($matches[1]);

      // Capture each requested attribute (if present)
      $collected = [];
      foreach ($attributes as $attr) {
        // Build a sub-pattern for e.g. \bsrc="([^"]*)".
        $attrPattern = '/\b' . preg_quote($attr, '/') . '="([^"]*)"/i';
        if (preg_match($attrPattern, $fullTag, $attrMatch)) {
          $collected[] = $attr . '="' . $attrMatch[1] . '"';
        }
      }

      // Build bracketed string, e.g. [IMG alt="..." title="..."].
      if ($collected) {
        return '[' . strtoupper($tagName) . ' ' . implode(' ', $collected) . ']';
      }
      else {
        return '[' . strtoupper($tagName) . ']';
      }
    }, $content);

    return empty($content) ? "" : $content;
  }

  /**
   * Add Markdown-style anchors for.
   *
   * @param string $html
   *   The HTML content to process.
   *
   * @return string|null
   *   The processed HTML content.
   */
  public function addMarkdownToHeadings($html) {
    if (empty($html)) {
      return NULL;
    }
    // Process from h1 to h6.
    for ($i = 1; $i <= 6; $i++) {
      if (empty($html)) {
        return NULL;
      }
      $html = preg_replace_callback(
        '/<h' . $i . '(.*?)>(.*?)<\/h' . $i . '>/is',
        function ($matches) use ($i) {
          $attrs = $matches[1];
          $content = trim($matches[2]);
          $hashes = str_repeat('#', $i);
          return "<h$i$attrs>$hashes $content</h$i>";
        },
        $html
      );
    }
    return $html;
  }

}
